from dagster import Array, BoolSource, Field, Noneable, Permissive, Shape, StringSource

# Shared between user code launcher config and container context
SHARED_ECS_CONFIG = {
    "server_resources": Field(
        Permissive(
            {
                "cpu": Field(
                    str,
                    is_required=False,
                    description="The CPU override to use for the launched task.",
                ),
                "memory": Field(
                    str,
                    is_required=False,
                    description="The memory override to use for the launched task.",
                ),
                "ephemeral_storage": Field(
                    int,
                    is_required=False,
                    description="The ephemeral storage, in GiB, to use for the launched task.",
                ),
                "repository_credentials": Field(
                    str,
                    is_required=False,
                    description=(
                        "AWS SSM Secret ARN of the credentials used to pull a server's image from a"
                        " private registry."
                        " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/private-auth.html"
                    ),
                ),
            }
        )
    ),
    "run_resources": Field(
        Permissive(
            {
                "cpu": Field(
                    str,
                    is_required=False,
                    description="The CPU override to use for the launched task.",
                ),
                "memory": Field(
                    str,
                    is_required=False,
                    description="The memory override to use for the launched task.",
                ),
                "ephemeral_storage": Field(
                    int,
                    is_required=False,
                    description="The ephemeral storage, in GiB, to use for the launched task.",
                ),
                "repository_credentials": Field(
                    str,
                    is_required=False,
                    description=(
                        "AWS SSM Secret ARN of the credentials used to pull a run's image from a"
                        " private registry."
                        " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/private-auth.html"
                    ),
                ),
            }
        )
    ),
    "runtime_platform": Field(
        Shape(
            {
                "cpuArchitecture": Field(StringSource, is_required=False),
                "operatingSystemFamily": Field(StringSource, is_required=False),
            }
        ),
        is_required=False,
        description=(
            "The operating system that the task definition is running on. See"
            " https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html#ECS.Client.register_task_definition"
            " for the available options."
        ),
    ),
    "volumes": Field(
        Array(
            Permissive(
                {
                    "name": Field(StringSource, is_required=False),
                }
            )
        ),
        is_required=False,
        description=(
            "List of data volume definitions for the task. See"
            " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/efs-volumes.html"
            " for the full list of available options."
        ),
    ),
    "mount_points": Field(
        Array(
            Shape(
                {
                    "sourceVolume": Field(StringSource, is_required=False),
                    "containerPath": Field(StringSource, is_required=False),
                    "readOnly": Field(BoolSource, is_required=False),
                }
            )
        ),
        is_required=False,
        description=(
            "Mount points for data volumes in the main container of the task."
            " See https://docs.aws.amazon.com/AmazonECS/latest/developerguide/efs-volumes.html"
            " for more information."
        ),
    ),
    "server_sidecar_containers": Field(
        Array(Permissive({})),
        is_required=False,
        description="Additional sidecar containers to include in code server task definitions.",
    ),
    "run_sidecar_containers": Field(
        Array(Permissive({})),
        is_required=False,
        description="Additional sidecar containers to include in run task definitions.",
    ),
    "run_ecs_tags": Field(
        Array(
            {
                "key": Field(StringSource, is_required=True),
                "value": Field(StringSource, is_required=False),
            }
        ),
        is_required=False,
        description="Additional tags to apply to the launched ECS task.",
    ),
    "server_ecs_tags": Field(
        Array(
            {
                "key": Field(StringSource, is_required=True),
                "value": Field(StringSource, is_required=False),
            }
        ),
        is_required=False,
        description="Additional tags to apply to the launched ECS task for a code server.",
    ),
}


ECS_CONTAINER_CONTEXT_CONFIG = {
    "secrets": Field(
        Noneable(Array(Shape({"name": StringSource, "valueFrom": StringSource}))),
        is_required=False,
        description=(
            "An array of AWS Secrets Manager secrets. These secrets will "
            "be mounted as environment variables in the container. See "
            "https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_Secret.html."
        ),
    ),
    "secrets_tags": Field(
        Noneable(Array(StringSource)),
        is_required=False,
        description=(
            "AWS Secrets Manager secrets with these tags will be mounted as "
            "environment variables in the container."
        ),
    ),
    "env_vars": Field(
        [StringSource],
        is_required=False,
        description=(
            "List of environment variable names to include in the ECS task. "
            "Each can be of the form KEY=VALUE or just KEY (in which case the value will be pulled "
            "from the current process)"
        ),
    ),
    "task_role_arn": Field(
        StringSource,
        is_required=False,
        description=(
            "ARN of the IAM role for launched tasks. See"
            " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-iam-roles.html. "
        ),
    ),
    "execution_role_arn": Field(
        StringSource,
        is_required=False,
        description=(
            "ARN of the task execution role for the ECS container and Fargate agent to make AWS API"
            " calls on your behalf. See"
            " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task_execution_IAM_role.html. "
        ),
    ),
    **SHARED_ECS_CONFIG,
}



import json
from collections import defaultdict
from os import getenv
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)
from uuid import uuid4

import dagster._check as check
from dagster import DagsterInvalidInvocationError
from dagster._core.assets import AssetDetails
from dagster._core.definitions.events import AssetKey, ExpectationResult
from dagster._core.event_api import EventLogRecord, EventRecordsFilter, RunShardedEventsCursor
from dagster._core.events import DagsterEvent, DagsterEventType
from dagster._core.events.log import EventLogEntry
from dagster._core.execution.stats import RunStepKeyStatsSnapshot, RunStepMarker, StepEventStatus
from dagster._core.storage.asset_check_execution_record import AssetCheckExecutionRecord
from dagster._core.storage.dagster_run import DagsterRunStatsSnapshot
from dagster._core.storage.event_log.base import (
    AssetEntry,
    AssetRecord,
    EventLogConnection,
    EventLogStorage,
)
from dagster._core.storage.partition_status_cache import AssetStatusCacheValue
from dagster._serdes import (
    ConfigurableClass,
    ConfigurableClassData,
    serialize_value,
)
from dagster._serdes.serdes import deserialize_value
from dagster._utils import datetime_as_float, utc_datetime_from_timestamp
from dagster._utils.concurrency import (
    ConcurrencyClaimStatus,
    ConcurrencyKeyInfo,
    ConcurrencySlotStatus,
)
from dagster._utils.error import SerializableErrorInfo
from dagster._utils.merger import merge_dicts
from dagster_cloud_cli.core.errors import GraphQLStorageError
from typing_extensions import Self

from dagster_cloud.storage.event_logs.utils import truncate_event

from .queries import (
    ADD_DYNAMIC_PARTITIONS_MUTATION,
    CHECK_CONCURRENCY_CLAIM_QUERY,
    CLAIM_CONCURRENCY_SLOT_MUTATION,
    DELETE_DYNAMIC_PARTITION_MUTATION,
    DELETE_EVENTS_MUTATION,
    ENABLE_SECONDARY_INDEX_MUTATION,
    FREE_CONCURRENCY_SLOT_FOR_STEP_MUTATION,
    FREE_CONCURRENCY_SLOTS_FOR_RUN_MUTATION,
    GET_ALL_ASSET_KEYS_QUERY,
    GET_ASSET_RECORDS_QUERY,
    GET_ASSET_RUN_IDS_QUERY,
    GET_CONCURRENCY_INFO_QUERY,
    GET_CONCURRENCY_KEYS_QUERY,
    GET_DYNAMIC_PARTITIONS_QUERY,
    GET_EVENT_RECORDS_QUERY,
    GET_EVENT_TAGS_FOR_ASSET,
    GET_LATEST_ASSET_PARTITION_MATERIALIZATION_ATTEMPTS_WITHOUT_MATERIALIZATIONS,
    GET_LATEST_MATERIALIZATION_EVENTS_QUERY,
    GET_LATEST_STORAGE_ID_BY_PARTITION,
    GET_LATEST_TAGS_BY_PARTITION,
    GET_MATERIALIZATION_COUNT_BY_PARTITION,
    GET_RECORDS_FOR_RUN_QUERY,
    GET_STATS_FOR_RUN_QUERY,
    GET_STEP_STATS_FOR_RUN_QUERY,
    HAS_ASSET_KEY_QUERY,
    HAS_DYNAMIC_PARTITION_QUERY,
    IS_ASSET_AWARE_QUERY,
    IS_PERSISTENT_QUERY,
    REINDEX_MUTATION,
    SET_CONCURRENCY_SLOTS_MUTATION,
    STORE_EVENT_MUTATION,
    UPDATE_ASSET_CACHED_STATUS_DATA_MUTATION,
    UPGRADE_EVENT_LOG_STORAGE_MUTATION,
    WIPE_ASSET_CACHED_STATUS_DATA_MUTATION,
    WIPE_ASSET_MUTATION,
    WIPE_EVENT_LOG_STORAGE_MUTATION,
)


def _input_for_serializable_error_info(serializable_error_info: Optional[SerializableErrorInfo]):
    check.opt_inst_param(serializable_error_info, "serializable_error_info", SerializableErrorInfo)

    if serializable_error_info is None:
        return None

    return {
        "message": serializable_error_info.message,
        "stack": serializable_error_info.stack,
        "clsName": serializable_error_info.cls_name,
        "cause": _input_for_serializable_error_info(serializable_error_info.cause),
    }


def _input_for_dagster_event(dagster_event: Optional[DagsterEvent]):
    if dagster_event is None:
        return None

    return {
        "eventTypeValue": dagster_event.event_type_value,
        "pipelineName": dagster_event.job_name,
        "stepHandleKey": dagster_event.step_handle.to_key() if dagster_event.step_handle else None,
        "solidHandleId": (
            dagster_event.node_handle.to_string() if dagster_event.node_handle else None
        ),
        "stepKindValue": dagster_event.step_kind_value,
        "loggingTags": (
            json.dumps(dagster_event.logging_tags) if dagster_event.logging_tags else None
        ),
        "eventSpecificData": (
            serialize_value(dagster_event.event_specific_data)
            if dagster_event.event_specific_data
            else None
        ),
        "message": dagster_event.message,
        "pid": dagster_event.pid,
        "stepKey": dagster_event.step_key,
    }


def _event_log_entry_from_graphql(graphene_event_log_entry: Dict) -> EventLogEntry:
    check.dict_param(graphene_event_log_entry, "graphene_event_log_entry")

    return EventLogEntry(
        error_info=(
            deserialize_value(
                check.str_elem(graphene_event_log_entry, "errorInfo"), SerializableErrorInfo
            )
            if graphene_event_log_entry.get("errorInfo") is not None
            else None
        ),
        level=check.int_elem(graphene_event_log_entry, "level"),
        user_message=check.str_elem(graphene_event_log_entry, "userMessage"),
        run_id=check.str_elem(graphene_event_log_entry, "runId"),
        timestamp=check.float_elem(graphene_event_log_entry, "timestamp"),
        step_key=check.opt_str_elem(graphene_event_log_entry, "stepKey"),
        job_name=check.opt_str_elem(graphene_event_log_entry, "pipelineName"),
        dagster_event=(
            deserialize_value(
                check.str_elem(graphene_event_log_entry, "dagsterEvent"),
                DagsterEvent,
            )
            if graphene_event_log_entry.get("dagsterEvent") is not None
            else None
        ),
    )


def _get_event_records_filter_input(
    event_records_filter,
) -> Optional[Dict[str, Any]]:
    check.opt_inst_param(event_records_filter, "event_records_filter", EventRecordsFilter)

    if event_records_filter is None:
        return None

    run_updated_timestamp = None
    if isinstance(event_records_filter.after_cursor, RunShardedEventsCursor):
        # we should parse this correctly, even if this is a vestigial field that only has semantic
        # meaning in the context of a run-sharded cursor against a SQLite backed event-log
        updated_dt = event_records_filter.after_cursor.run_updated_after
        run_updated_timestamp = (
            updated_dt.timestamp() if updated_dt.tzinfo else datetime_as_float(updated_dt)
        )

    return {
        "eventType": (
            event_records_filter.event_type.value if event_records_filter.event_type else None
        ),
        "assetKey": (
            event_records_filter.asset_key.to_string() if event_records_filter.asset_key else None
        ),
        "assetPartitions": event_records_filter.asset_partitions,
        "afterTimestamp": event_records_filter.after_timestamp,
        "beforeTimestamp": event_records_filter.before_timestamp,
        "beforeCursor": (
            event_records_filter.before_cursor.id
            if isinstance(event_records_filter.before_cursor, RunShardedEventsCursor)
            else event_records_filter.before_cursor
        ),
        "afterCursor": (
            event_records_filter.after_cursor.id
            if isinstance(event_records_filter.after_cursor, RunShardedEventsCursor)
            else event_records_filter.after_cursor
        ),
        "runUpdatedAfter": run_updated_timestamp,
        "tags": (
            [
                merge_dicts(
                    {"key": tag_key},
                    ({"value": tag_value} if isinstance(tag_value, str) else {"values": tag_value}),
                )
                for tag_key, tag_value in event_records_filter.tags.items()
            ]
            if event_records_filter.tags
            else None
        ),
        "storageIds": event_records_filter.storage_ids,
    }


def _event_record_from_graphql(graphene_event_record: Dict) -> EventLogRecord:
    check.dict_param(graphene_event_record, "graphene_event_record")

    return EventLogRecord(
        storage_id=check.int_elem(graphene_event_record, "storageId"),
        event_log_entry=_event_log_entry_from_graphql(graphene_event_record["eventLogEntry"]),
    )


def _asset_entry_from_graphql(graphene_asset_entry: Dict) -> AssetEntry:
    check.dict_param(graphene_asset_entry, "graphene_asset_entry")
    return AssetEntry(
        asset_key=AssetKey(graphene_asset_entry["assetKey"]["path"]),
        last_materialization_record=(
            _event_record_from_graphql(graphene_asset_entry["lastMaterializationRecord"])
            if graphene_asset_entry["lastMaterializationRecord"]
            else None
        ),
        last_run_id=graphene_asset_entry["lastRunId"],
        asset_details=(
            AssetDetails(
                last_wipe_timestamp=graphene_asset_entry["assetDetails"]["lastWipeTimestamp"]
            )
            if graphene_asset_entry["assetDetails"]
            else None
        ),
        cached_status=(
            AssetStatusCacheValue(
                latest_storage_id=graphene_asset_entry["cachedStatus"]["latestStorageId"],
                partitions_def_id=graphene_asset_entry["cachedStatus"]["partitionsDefId"],
                serialized_materialized_partition_subset=graphene_asset_entry["cachedStatus"][
                    "serializedMaterializedPartitionSubset"
                ],
                serialized_failed_partition_subset=graphene_asset_entry["cachedStatus"][
                    "serializedFailedPartitionSubset"
                ],
                serialized_in_progress_partition_subset=graphene_asset_entry["cachedStatus"][
                    "serializedInProgressPartitionSubset"
                ],
                earliest_in_progress_materialization_event_id=graphene_asset_entry["cachedStatus"][
                    "earliestInProgressMaterializationEventId"
                ],
            )
            if graphene_asset_entry["cachedStatus"]
            else None
        ),
    )


def _asset_record_from_graphql(graphene_asset_record: Dict) -> AssetRecord:
    check.dict_param(graphene_asset_record, "graphene_asset_record")
    return AssetRecord(
        storage_id=graphene_asset_record["storageId"],
        asset_entry=_asset_entry_from_graphql(graphene_asset_record["assetEntry"]),
    )


class GraphQLEventLogStorage(EventLogStorage, ConfigurableClass):
    def __init__(self, inst_data=None, override_graphql_client=None):
        """Initialize this class directly only for test. Use the ConfigurableClass machinery to
        init from instance yaml.
        """
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._override_graphql_client = override_graphql_client

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return GraphQLEventLogStorage(inst_data=inst_data)

    @property
    def _graphql_client(self):
        return (
            self._override_graphql_client
            if self._override_graphql_client
            else self._instance.graphql_client
        )

    def _execute_query(self, query, variables=None, headers=None):
        res = self._graphql_client.execute(
            query,
            variable_values=variables,
            headers=headers,
        )
        if "errors" in res:
            raise GraphQLStorageError(res)
        return res

    def get_records_for_run(
        self,
        run_id: str,
        cursor: Optional[str] = None,
        of_type: Optional[Union[DagsterEventType, Set[DagsterEventType]]] = None,
        limit: Optional[int] = None,
        ascending: bool = True,
    ) -> EventLogConnection:
        check.invariant(not of_type or isinstance(of_type, (DagsterEventType, frozenset, set)))

        is_of_type_set = isinstance(of_type, (set, frozenset))

        res = self._execute_query(
            GET_RECORDS_FOR_RUN_QUERY,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "cursor": check.opt_str_param(cursor, "cursor"),
                "ofType": (
                    cast(DagsterEventType, of_type).value
                    if of_type and not is_of_type_set
                    else None
                ),
                "ofTypes": (
                    [dagster_type.value for dagster_type in cast(set, of_type)]
                    if of_type and is_of_type_set
                    else None
                ),
                "limit": limit,
                # only send False for back-compat since True is the default
                **({"ascending": False} if not ascending else {}),
            },
        )
        connection_data = res["data"]["eventLogs"]["getRecordsForRun"]
        return EventLogConnection(
            records=[_event_record_from_graphql(record) for record in connection_data["records"]],
            cursor=connection_data["cursor"],
            has_more=connection_data["hasMore"],
        )

    def get_stats_for_run(self, run_id: str) -> DagsterRunStatsSnapshot:
        res = self._execute_query(
            GET_STATS_FOR_RUN_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        stats = res["data"]["eventLogs"]["getStatsForRun"]
        return DagsterRunStatsSnapshot(
            run_id=check.str_elem(stats, "runId"),
            steps_succeeded=check.int_elem(stats, "stepsSucceeded"),
            steps_failed=check.int_elem(stats, "stepsFailed"),
            materializations=check.int_elem(stats, "materializations"),
            expectations=check.int_elem(stats, "expectations"),
            enqueued_time=check.opt_float_elem(stats, "enqueuedTime"),
            launch_time=check.opt_float_elem(stats, "launchTime"),
            start_time=check.opt_float_elem(stats, "startTime"),
            end_time=check.opt_float_elem(stats, "endTime"),
        )

    def get_step_stats_for_run(
        self, run_id: str, step_keys: Optional[List[str]] = None
    ) -> List[RunStepKeyStatsSnapshot]:
        res = self._execute_query(
            GET_STEP_STATS_FOR_RUN_QUERY,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "stepKeys": check.opt_list_param(step_keys, "step_keys", of_type=str),
            },
        )
        step_stats = res["data"]["eventLogs"]["getStepStatsForRun"]
        return [
            RunStepKeyStatsSnapshot(
                run_id=check.str_elem(stats, "runId"),
                step_key=check.str_elem(stats, "stepKey"),
                status=(
                    getattr(StepEventStatus, check.str_elem(stats, "status"))
                    if stats.get("status") is not None
                    else None
                ),
                start_time=check.opt_float_elem(stats, "startTime"),
                end_time=check.opt_float_elem(stats, "endTime"),
                materialization_events=[
                    deserialize_value(materialization_event, EventLogEntry)
                    for materialization_event in check.opt_list_elem(
                        stats, "materializationEvents", of_type=str
                    )
                ],
                expectation_results=[
                    deserialize_value(expectation_result, ExpectationResult)
                    for expectation_result in check.opt_list_elem(
                        stats, "expectationResults", of_type=str
                    )
                ],
                attempts=check.opt_int_elem(stats, "attempts"),
                attempts_list=[
                    deserialize_value(marker, RunStepMarker)
                    for marker in check.opt_list_elem(stats, "attemptsList", of_type=str)
                ],
                markers=[
                    deserialize_value(marker, RunStepMarker)
                    for marker in check.opt_list_elem(stats, "markers", of_type=str)
                ],
            )
            for stats in step_stats
        ]

    def store_event(self, event: EventLogEntry):
        check.inst_param(event, "event", EventLogEntry)

        event = truncate_event(event)

        headers = {}
        # env var opt-in for experimental feature
        if getenv("DAGSTER_CLOUD_IDEMPOTENT_STORE_EVENT"):
            headers = {"Idempotency-Key": str(uuid4())}

        self._execute_query(
            STORE_EVENT_MUTATION,
            variables={
                "eventRecord": {
                    "errorInfo": _input_for_serializable_error_info(event.error_info),
                    "level": event.level,
                    "userMessage": event.user_message,
                    "runId": event.run_id,
                    "timestamp": event.timestamp,
                    "stepKey": event.step_key,
                    "pipelineName": event.job_name,
                    "dagsterEvent": _input_for_dagster_event(event.dagster_event),
                }
            },
            headers=headers,
        )

    def delete_events(self, run_id: str):
        self._execute_query(
            DELETE_EVENTS_MUTATION, variables={"runId": check.str_param(run_id, "run_id")}
        )
        return

    def upgrade(self):
        return self._execute_query(UPGRADE_EVENT_LOG_STORAGE_MUTATION)

    def reindex_assets(self, print_fn: Optional[Callable] = lambda _: None, force: bool = False):
        return self._execute_query(
            REINDEX_MUTATION, variables={"force": check.bool_param(force, "force")}
        )

    def reindex_events(self, print_fn: Optional[Callable] = lambda _: None, force: bool = False):
        return self._execute_query(
            REINDEX_MUTATION, variables={"force": check.bool_param(force, "force")}
        )

    def wipe(self):
        return self._execute_query(WIPE_EVENT_LOG_STORAGE_MUTATION)

    def watch(self, run_id: str, cursor: str, callback: Callable):
        raise NotImplementedError("Not callable from user cloud")

    def end_watch(self, run_id: str, handler: Callable):
        raise NotImplementedError("Not callable from user cloud")

    def enable_secondary_index(self, name: str, run_id: Optional[str] = None):
        return self._execute_query(
            ENABLE_SECONDARY_INDEX_MUTATION,
            variables={
                "name": check.str_param(name, "name"),
                "runId": check.str_param(run_id, "run_id"),
            },
        )

    @property
    def is_persistent(self) -> bool:
        res = self._execute_query(IS_PERSISTENT_QUERY)
        return res["data"]["eventLogs"]["isPersistent"]

    @property
    def is_asset_aware(self) -> bool:
        res = self._execute_query(IS_ASSET_AWARE_QUERY)
        return res["data"]["eventLogs"]["isAssetAware"]

    def get_event_records(
        self,
        event_records_filter: Optional[EventRecordsFilter] = None,
        limit: Optional[int] = None,
        ascending: Optional[bool] = False,
    ) -> Iterable[EventLogRecord]:
        check.opt_inst_param(event_records_filter, "event_records_filter", EventRecordsFilter)
        check.opt_int_param(limit, "limit")
        check.bool_param(ascending, "ascending")

        res = self._execute_query(
            GET_EVENT_RECORDS_QUERY,
            variables={
                "eventRecordsFilter": _get_event_records_filter_input(event_records_filter),
                "limit": limit,
                "ascending": ascending,
            },
        )
        return [
            _event_record_from_graphql(result)
            for result in res["data"]["eventLogs"]["getEventRecords"]
        ]

    def get_asset_records(
        self, asset_keys: Optional[Sequence[AssetKey]] = None
    ) -> Iterable[AssetRecord]:
        res = self._execute_query(
            GET_ASSET_RECORDS_QUERY,
            variables={
                "assetKeys": (
                    [asset_key.to_string() for asset_key in asset_keys] if asset_keys else []
                )
            },
        )

        return [
            _asset_record_from_graphql(result)
            for result in res["data"]["eventLogs"]["getAssetRecords"]
        ]

    def has_asset_key(self, asset_key: AssetKey):
        check.inst_param(asset_key, "asset_key", AssetKey)

        res = self._execute_query(
            HAS_ASSET_KEY_QUERY, variables={"assetKey": asset_key.to_string()}
        )
        return res["data"]["eventLogs"]["hasAssetKey"]

    def all_asset_keys(self) -> Iterable[AssetKey]:
        res = self._execute_query(GET_ALL_ASSET_KEYS_QUERY)
        return [
            check.not_none(AssetKey.from_db_string(asset_key_string))
            for asset_key_string in res["data"]["eventLogs"]["getAllAssetKeys"]
        ]

    def get_latest_materialization_events(
        self, asset_keys: Sequence[AssetKey]
    ) -> Mapping[AssetKey, Optional[EventLogEntry]]:
        check.list_param(asset_keys, "asset_keys", of_type=AssetKey)

        res = self._execute_query(
            GET_LATEST_MATERIALIZATION_EVENTS_QUERY,
            variables={"assetKeys": [asset_key.to_string() for asset_key in asset_keys]},
        )

        result = {}
        for entry in res["data"]["eventLogs"]["getLatestMaterializationEvents"]:
            event = _event_log_entry_from_graphql(entry)
            if event.dagster_event and event.dagster_event.asset_key is not None:
                result[event.dagster_event.asset_key] = event

        return result

    def get_asset_run_ids(self, asset_key: AssetKey) -> Iterable[str]:
        res = self._execute_query(
            GET_ASSET_RUN_IDS_QUERY, variables={"assetKey": asset_key.to_string()}
        )
        return res["data"]["eventLogs"]["getAssetRunIds"]

    def can_cache_asset_status_data(self) -> bool:
        # cached_status_data column exists in the asset_key table
        return True

    def update_asset_cached_status_data(
        self, asset_key: AssetKey, cache_values: AssetStatusCacheValue
    ) -> None:
        self._execute_query(
            UPDATE_ASSET_CACHED_STATUS_DATA_MUTATION,
            variables={
                "assetKey": asset_key.to_string(),
                "cacheValues": {
                    "latestStorageId": cache_values.latest_storage_id,
                    "partitionsDefId": cache_values.partitions_def_id,
                    "serializedMaterializedPartitionSubset": (
                        cache_values.serialized_materialized_partition_subset
                    ),
                    "serializedFailedPartitionSubset": (
                        cache_values.serialized_failed_partition_subset
                    ),
                    "serializedInProgressPartitionSubset": (
                        cache_values.serialized_in_progress_partition_subset
                    ),
                    "earliestInProgressMaterializationEventId": (
                        cache_values.earliest_in_progress_materialization_event_id
                    ),
                },
            },
        )

    def wipe_asset_cached_status(self, asset_key: AssetKey) -> None:
        res = self._execute_query(
            WIPE_ASSET_CACHED_STATUS_DATA_MUTATION, variables={"assetKey": asset_key.to_string()}
        )
        return res

    def get_materialization_count_by_partition(
        self,
        asset_keys: Sequence[AssetKey],
        after_cursor: Optional[int] = None,
    ) -> Mapping[AssetKey, Mapping[str, int]]:
        check.list_param(asset_keys, "asset_keys", of_type=AssetKey)
        check.opt_int_param(after_cursor, "after_cursor")

        res = self._execute_query(
            GET_MATERIALIZATION_COUNT_BY_PARTITION,
            variables={
                "assetKeys": [asset_key.to_string() for asset_key in asset_keys],
                "afterCursor": after_cursor,
            },
        )

        materialization_count_result = res["data"]["eventLogs"][
            "getMaterializationCountByPartition"
        ]

        materialization_count_by_partition: Dict[AssetKey, Dict[str, int]] = {
            asset_key: {} for asset_key in asset_keys
        }
        for asset_count in materialization_count_result:
            asset_key = check.not_none(AssetKey.from_db_string(asset_count["assetKey"]))
            for graphene_partition_count in asset_count["materializationCountByPartition"]:
                materialization_count_by_partition[asset_key][
                    graphene_partition_count["partition"]
                ] = graphene_partition_count["materializationCount"]

        return materialization_count_by_partition

    def get_latest_storage_id_by_partition(
        self, asset_key: AssetKey, event_type: DagsterEventType
    ) -> Mapping[str, int]:
        res = self._execute_query(
            GET_LATEST_STORAGE_ID_BY_PARTITION,
            variables={
                "assetKey": asset_key.to_string(),
                "eventType": event_type.value,
            },
        )
        latest_storage_id_result = res["data"]["eventLogs"]["getLatestStorageIdByPartition"]
        latest_storage_id_by_partition: Dict[str, int] = {}

        for graphene_latest_storage_id in latest_storage_id_result:
            latest_storage_id_by_partition[graphene_latest_storage_id["partition"]] = (
                graphene_latest_storage_id["storageId"]
            )

        return latest_storage_id_by_partition

    def get_latest_tags_by_partition(
        self,
        asset_key: AssetKey,
        event_type: DagsterEventType,
        tag_keys: Optional[Sequence[str]] = None,
        asset_partitions: Optional[Sequence[str]] = None,
        before_cursor: Optional[int] = None,
        after_cursor: Optional[int] = None,
    ) -> Mapping[str, Mapping[str, str]]:
        res = self._execute_query(
            GET_LATEST_TAGS_BY_PARTITION,
            variables={
                "assetKey": asset_key.to_string(),
                "eventType": event_type.value,
                "tagKeys": tag_keys,
                "assetPartitions": asset_partitions,
                "beforeCursor": before_cursor,
                "afterCursor": after_cursor,
            },
        )
        latest_tags_by_partition_result = res["data"]["eventLogs"]["getLatestTagsByPartition"]
        latest_tags_by_partition: Dict[str, Dict[str, str]] = defaultdict(dict)
        for tag_by_partition in latest_tags_by_partition_result:
            latest_tags_by_partition[tag_by_partition["partition"]][tag_by_partition["key"]] = (
                tag_by_partition["value"]
            )

        # convert defaultdict to dict
        return dict(latest_tags_by_partition)

    def get_latest_asset_partition_materialization_attempts_without_materializations(
        self, asset_key: AssetKey
    ) -> Mapping[str, Tuple[str, int]]:
        res = self._execute_query(
            GET_LATEST_ASSET_PARTITION_MATERIALIZATION_ATTEMPTS_WITHOUT_MATERIALIZATIONS,
            variables={
                "assetKey": asset_key.to_string(),
            },
        )
        return res["data"]["eventLogs"][
            "getLatestAssetPartitionMaterializationAttemptsWithoutMaterializations"
        ]

    def get_event_tags_for_asset(
        self,
        asset_key: AssetKey,
        filter_tags: Optional[Mapping[str, str]] = None,
        filter_event_id: Optional[int] = None,
    ) -> Sequence[Mapping[str, str]]:
        check.inst_param(asset_key, "asset_key", AssetKey)
        filter_tags = check.opt_mapping_param(
            filter_tags, "filter_tags", key_type=str, value_type=str
        )

        res = self._execute_query(
            GET_EVENT_TAGS_FOR_ASSET,
            variables={
                "assetKey": asset_key.to_string(),
                "filterTags": [{"key": key, "value": value} for key, value in filter_tags.items()],
                "filterEventId": filter_event_id,
            },
        )
        tags_result = res["data"]["eventLogs"]["getAssetEventTags"]
        return [
            {event_tag["key"]: event_tag["value"] for event_tag in event_tags["tags"]}
            for event_tags in tags_result
        ]

    def get_dynamic_partitions(self, partitions_def_name: str) -> Sequence[str]:
        check.str_param(partitions_def_name, "partitions_def_name")
        res = self._execute_query(
            GET_DYNAMIC_PARTITIONS_QUERY,
            variables={"partitionsDefName": partitions_def_name},
        )
        return res["data"]["eventLogs"]["getDynamicPartitions"]

    def has_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> bool:
        check.str_param(partitions_def_name, "partitions_def_name")
        check.str_param(partition_key, "partition_key")
        res = self._execute_query(
            HAS_DYNAMIC_PARTITION_QUERY,
            variables={
                "partitionsDefName": partitions_def_name,
                "partitionKey": partition_key,
            },
        )
        return res["data"]["eventLogs"]["hasDynamicPartition"]

    def add_dynamic_partitions(
        self, partitions_def_name: str, partition_keys: Sequence[str]
    ) -> None:
        check.str_param(partitions_def_name, "partitions_def_name")
        check.sequence_param(partition_keys, "partition_keys", of_type=str)
        self._execute_query(
            ADD_DYNAMIC_PARTITIONS_MUTATION,
            variables={
                "partitionsDefName": partitions_def_name,
                "partitionKeys": partition_keys,
            },
        )

    def delete_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> None:
        check.str_param(partitions_def_name, "partitions_def_name")
        check.str_param(partition_key, "partition_key")
        self._execute_query(
            DELETE_DYNAMIC_PARTITION_MUTATION,
            variables={
                "partitionsDefName": partitions_def_name,
                "partitionKey": partition_key,
            },
        )

    def wipe_asset(self, asset_key: AssetKey):
        res = self._execute_query(
            WIPE_ASSET_MUTATION, variables={"assetKey": asset_key.to_string()}
        )
        return res

    @property
    def supports_global_concurrency_limits(self) -> bool:
        return True

    def set_concurrency_slots(self, concurrency_key: str, num: int) -> None:
        check.str_param(concurrency_key, "concurrency_key")
        check.int_param(num, "num")
        res = self._execute_query(
            SET_CONCURRENCY_SLOTS_MUTATION,
            variables={"concurrencyKey": concurrency_key, "num": num},
        )
        result = res["data"]["eventLogs"]["SetConcurrencySlots"]
        error = result.get("error")

        if error:
            if error["className"] == "DagsterInvalidInvocationError":
                raise DagsterInvalidInvocationError(error["message"])
            else:
                raise GraphQLStorageError(res)
        return res

    def get_concurrency_keys(self) -> Set[str]:
        res = self._execute_query(GET_CONCURRENCY_KEYS_QUERY)
        return set(res["data"]["eventLogs"]["getConcurrencyKeys"])

    def get_concurrency_info(self, concurrency_key: str) -> ConcurrencyKeyInfo:
        check.str_param(concurrency_key, "concurrency_key")
        res = self._execute_query(
            GET_CONCURRENCY_INFO_QUERY,
            variables={"concurrencyKey": concurrency_key},
        )
        info = res["data"]["eventLogs"]["getConcurrencyInfo"]
        return ConcurrencyKeyInfo(
            concurrency_key=concurrency_key,
            slot_count=info["slotCount"],
            active_slot_count=info["activeSlotCount"],
            active_run_ids=set(info["activeRunIds"]),
            pending_step_count=info["pendingStepCount"],
            pending_run_ids=set(info["pendingStepRunIds"]),
            assigned_step_count=info["assignedStepCount"],
            assigned_run_ids=set(info["assignedStepRunIds"]),
        )

    def claim_concurrency_slot(
        self, concurrency_key: str, run_id: str, step_key: str, priority: Optional[int] = None
    ) -> ConcurrencyClaimStatus:
        check.str_param(concurrency_key, "concurrency_key")
        check.str_param(run_id, "run_id")
        check.str_param(step_key, "step_key")
        check.opt_int_param(priority, "priority")
        res = self._execute_query(
            CLAIM_CONCURRENCY_SLOT_MUTATION,
            variables={
                "concurrencyKey": concurrency_key,
                "runId": run_id,
                "stepKey": step_key,
                "priority": priority,
            },
        )
        claim_status = res["data"]["eventLogs"]["ClaimConcurrencySlot"]["status"]
        return ConcurrencyClaimStatus(
            concurrency_key=concurrency_key,
            slot_status=ConcurrencySlotStatus(claim_status["slotStatus"]),
            priority=claim_status["priority"],
            assigned_timestamp=(
                utc_datetime_from_timestamp(claim_status["assignedTimestamp"])
                if claim_status["assignedTimestamp"]
                else None
            ),
            enqueued_timestamp=(
                utc_datetime_from_timestamp(claim_status["enqueuedTimestamp"])
                if claim_status["enqueuedTimestamp"]
                else None
            ),
        )

    def check_concurrency_claim(
        self, concurrency_key: str, run_id: str, step_key: str
    ) -> ConcurrencyClaimStatus:
        check.str_param(concurrency_key, "concurrency_key")
        check.str_param(run_id, "run_id")
        check.str_param(step_key, "step_key")
        res = self._execute_query(
            CHECK_CONCURRENCY_CLAIM_QUERY,
            variables={
                "concurrencyKey": concurrency_key,
                "runId": run_id,
                "stepKey": step_key,
            },
        )
        claim_status = res["data"]["eventLogs"]["getCheckConcurrencyClaim"]
        return ConcurrencyClaimStatus(
            concurrency_key=concurrency_key,
            slot_status=ConcurrencySlotStatus(claim_status["slotStatus"]),
            priority=claim_status["priority"],
            assigned_timestamp=(
                utc_datetime_from_timestamp(claim_status["assignedTimestamp"])
                if claim_status["assignedTimestamp"]
                else None
            ),
            enqueued_timestamp=(
                utc_datetime_from_timestamp(claim_status["enqueuedTimestamp"])
                if claim_status["enqueuedTimestamp"]
                else None
            ),
        )

    def get_concurrency_run_ids(self) -> Set[str]:
        raise NotImplementedError("Not callable from user cloud")

    def free_concurrency_slots_for_run(self, run_id: str) -> None:
        check.str_param(run_id, "run_id")
        res = self._execute_query(
            FREE_CONCURRENCY_SLOTS_FOR_RUN_MUTATION, variables={"runId": run_id}
        )
        return res

    def free_concurrency_slot_for_step(self, run_id: str, step_key: str) -> None:
        check.str_param(run_id, "run_id")
        check.opt_str_param(step_key, "step_key")
        res = self._execute_query(
            FREE_CONCURRENCY_SLOT_FOR_STEP_MUTATION,
            variables={"runId": run_id, "stepKey": step_key},
        )
        return res

    def get_asset_check_executions(
        self,
        asset_key: AssetKey,
        check_name: str,
        limit: int,
        cursor: Optional[int] = None,
        materialization_event_storage_id: Optional[int] = None,
        include_planned: bool = True,
    ) -> Sequence[AssetCheckExecutionRecord]:
        raise NotImplementedError("Not callable from user cloud")



# @generated

# This file was generated by running `python -m dagster_cloud.pex.grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output


# ruff: noqa: SLF001

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: multi_pex_api.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13multi_pex_api.proto\x12\rmulti_pex_api"\x07\n\x05\x45mpty"8\n\x16\x43reatePexServerRequest\x12\x1e\n\x16\x63reate_pex_server_args\x18\x01'
    b' \x01(\t":\n\x14\x43reatePexServerReply\x12"\n\x1a\x63reate_pex_server_response\x18\x01'
    b' \x01(\t"4\n\x14GetPexServersRequest\x12\x1c\n\x14get_pex_servers_args\x18\x01'
    b' \x01(\t"6\n\x12GetPexServersReply\x12 \n\x18get_pex_servers_response\x18\x01'
    b' \x01(\t"<\n\x18ShutdownPexServerRequest\x12 \n\x18shutdown_pex_server_args\x18\x01'
    b' \x01(\t">\n\x16ShutdownPexServerReply\x12$\n\x1cshutdown_pex_server_response\x18\x01'
    b' \x01(\t"\x1b\n\x0bPingRequest\x12\x0c\n\x04\x65\x63ho\x18\x01'
    b' \x01(\t"\x19\n\tPingReply\x12\x0c\n\x04\x65\x63ho\x18\x01'
    b' \x01(\t2\xf0\x02\n\x0bMultiPexApi\x12_\n\x0f\x43reatePexServer\x12%.multi_pex_api.CreatePexServerRequest\x1a#.multi_pex_api.CreatePexServerReply"\x00\x12Y\n\rGetPexServers\x12#.multi_pex_api.GetPexServersRequest\x1a!.multi_pex_api.GetPexServersReply"\x00\x12\x65\n\x11ShutdownPexServer\x12\'.multi_pex_api.ShutdownPexServerRequest\x1a%.multi_pex_api.ShutdownPexServerReply"\x00\x12>\n\x04Ping\x12\x1a.multi_pex_api.PingRequest\x1a\x18.multi_pex_api.PingReply"\x00\x62\x06proto3'
)


_EMPTY = DESCRIPTOR.message_types_by_name["Empty"]
_CREATEPEXSERVERREQUEST = DESCRIPTOR.message_types_by_name["CreatePexServerRequest"]
_CREATEPEXSERVERREPLY = DESCRIPTOR.message_types_by_name["CreatePexServerReply"]
_GETPEXSERVERSREQUEST = DESCRIPTOR.message_types_by_name["GetPexServersRequest"]
_GETPEXSERVERSREPLY = DESCRIPTOR.message_types_by_name["GetPexServersReply"]
_SHUTDOWNPEXSERVERREQUEST = DESCRIPTOR.message_types_by_name["ShutdownPexServerRequest"]
_SHUTDOWNPEXSERVERREPLY = DESCRIPTOR.message_types_by_name["ShutdownPexServerReply"]
_PINGREQUEST = DESCRIPTOR.message_types_by_name["PingRequest"]
_PINGREPLY = DESCRIPTOR.message_types_by_name["PingReply"]
Empty = _reflection.GeneratedProtocolMessageType(
    "Empty",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.Empty)
    },
)
_sym_db.RegisterMessage(Empty)

CreatePexServerRequest = _reflection.GeneratedProtocolMessageType(
    "CreatePexServerRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEPEXSERVERREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.CreatePexServerRequest)
    },
)
_sym_db.RegisterMessage(CreatePexServerRequest)

CreatePexServerReply = _reflection.GeneratedProtocolMessageType(
    "CreatePexServerReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEPEXSERVERREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.CreatePexServerReply)
    },
)
_sym_db.RegisterMessage(CreatePexServerReply)

GetPexServersRequest = _reflection.GeneratedProtocolMessageType(
    "GetPexServersRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPEXSERVERSREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.GetPexServersRequest)
    },
)
_sym_db.RegisterMessage(GetPexServersRequest)

GetPexServersReply = _reflection.GeneratedProtocolMessageType(
    "GetPexServersReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPEXSERVERSREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.GetPexServersReply)
    },
)
_sym_db.RegisterMessage(GetPexServersReply)

ShutdownPexServerRequest = _reflection.GeneratedProtocolMessageType(
    "ShutdownPexServerRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SHUTDOWNPEXSERVERREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.ShutdownPexServerRequest)
    },
)
_sym_db.RegisterMessage(ShutdownPexServerRequest)

ShutdownPexServerReply = _reflection.GeneratedProtocolMessageType(
    "ShutdownPexServerReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _SHUTDOWNPEXSERVERREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.ShutdownPexServerReply)
    },
)
_sym_db.RegisterMessage(ShutdownPexServerReply)

PingRequest = _reflection.GeneratedProtocolMessageType(
    "PingRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _PINGREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.PingRequest)
    },
)
_sym_db.RegisterMessage(PingRequest)

PingReply = _reflection.GeneratedProtocolMessageType(
    "PingReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _PINGREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.PingReply)
    },
)
_sym_db.RegisterMessage(PingReply)

_MULTIPEXAPI = DESCRIPTOR.services_by_name["MultiPexApi"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _EMPTY._serialized_start = 38
    _EMPTY._serialized_end = 45
    _CREATEPEXSERVERREQUEST._serialized_start = 47
    _CREATEPEXSERVERREQUEST._serialized_end = 103
    _CREATEPEXSERVERREPLY._serialized_start = 105
    _CREATEPEXSERVERREPLY._serialized_end = 163
    _GETPEXSERVERSREQUEST._serialized_start = 165
    _GETPEXSERVERSREQUEST._serialized_end = 217
    _GETPEXSERVERSREPLY._serialized_start = 219
    _GETPEXSERVERSREPLY._serialized_end = 273
    _SHUTDOWNPEXSERVERREQUEST._serialized_start = 275
    _SHUTDOWNPEXSERVERREQUEST._serialized_end = 335
    _SHUTDOWNPEXSERVERREPLY._serialized_start = 337
    _SHUTDOWNPEXSERVERREPLY._serialized_end = 399
    _PINGREQUEST._serialized_start = 401
    _PINGREQUEST._serialized_end = 428
    _PINGREPLY._serialized_start = 430
    _PINGREPLY._serialized_end = 455
    _MULTIPEXAPI._serialized_start = 458
    _MULTIPEXAPI._serialized_end = 826
# @@protoc_insertion_point(module_scope)



import dagster._check as check
from dagster_aws.ecs import EcsRunLauncher

from dagster_cloud.instance import DagsterCloudAgentInstance

from .utils import get_run_task_definition_family


class CloudEcsRunLauncher(EcsRunLauncher[DagsterCloudAgentInstance]):
    def _get_run_task_definition_family(self, run) -> str:
        return get_run_task_definition_family(
            self._instance.organization_name,
            check.not_none(self._instance.deployment_name),
            check.not_none(run.external_job_origin),
        )



"""Decodes PexMetadata to download S3 files locally and setup a runnable PEX environment."""
import hashlib
import logging
import os
import shutil
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional
from uuid import uuid4

from dagster import _check as check
from dagster_cloud_cli.core.workspace import PexMetadata

DEFAULT_PEX_FILES_DIR = "/tmp/pex-files"


def _download_from_s3(filename: str, local_filepath: str):
    # Lazy import boto3 to avoid a hard dependency during module load
    import boto3

    s3 = boto3.client("s3", region_name="us-west-2")

    # TODO: move the bucket and prefix to pex_metdata
    s3_bucket_name = os.environ["DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_BUCKET"]
    # prefix is typically org-storage/{org_public_id}
    s3_prefix = os.environ["DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_PREFIX"]

    s3_key = f"{s3_prefix}/pex/{filename}"
    # write to different file and rename for read safety
    local_tmp_filepath = local_filepath + "-" + str(uuid4())
    s3.download_file(
        Bucket=s3_bucket_name,
        Key=s3_key,
        Filename=local_tmp_filepath,
    )
    os.rename(local_tmp_filepath, local_filepath)


class PexExecutable(
    NamedTuple(
        "_PexExecutable",
        [
            ("source_path", str),
            ("all_paths", List[str]),
            ("environ", Dict[str, str]),
            ("working_directory", Optional[str]),
            ("venv_dirs", List[str]),
        ],
    )
):
    def __new__(
        cls,
        source_path: str,
        all_paths: List[str],
        environ: Dict[str, str],
        working_directory: Optional[str],
        venv_dirs: List[str],
    ):
        return super(PexExecutable, cls).__new__(
            cls,
            check.str_param(source_path, "source_path"),
            check.list_param(all_paths, "all_paths", str),
            check.dict_param(environ, "environ", str, str),
            check.opt_str_param(working_directory, "working_directory"),
            check.list_param(venv_dirs, "venv_dirs"),
        )


@dataclass
class PexVenv:
    path: Path
    site_packages: Path
    bin: Path
    entrypoint: Path
    pex_filename: str


class PexInstallationError(Exception):
    pass


class PexS3Registry:
    def __init__(self, local_pex_files_dir: Optional[str] = None):
        self._local_pex_files_dir = (
            local_pex_files_dir if local_pex_files_dir else DEFAULT_PEX_FILES_DIR
        )
        os.makedirs(self._local_pex_files_dir, exist_ok=True)
        self.working_dirs: Dict[str, str] = (
            {}
        )  # once unpacked, working dirs dont change so we cache them

    def get_pex_executable(self, pex_metadata: PexMetadata) -> PexExecutable:
        if "=" not in pex_metadata.pex_tag:
            raise ValueError(f"Invalid pex tag, no prefix in {pex_metadata.pex_tag!r}")
        prefix, filenames = pex_metadata.pex_tag.split("=", 1)
        if prefix != "files":
            raise ValueError(
                f'Expected pex_tag prefix "files=" not found in tag {pex_metadata.pex_tag!r}'
            )

        pex_filenames = filenames.split(":")
        deps_pex_filepaths = []
        source_pex_filepath = None

        for filename in pex_filenames:
            local_filepath = os.path.join(self._local_pex_files_dir, filename)

            # no need to download if we already have this file - these
            # files have a content hash suffix so name equality implies content is same
            if not os.path.exists(local_filepath):
                if os.getenv("S3_PEX_DISABLED"):
                    raise ValueError(
                        f"File {local_filepath} not found for pex tag {pex_metadata.pex_tag},"
                        " S3_PEX_DISABLED"
                    )
                _download_from_s3(filename, local_filepath)

            if filename.startswith("source-"):
                source_pex_filepath = local_filepath
                # make it executable
                os.chmod(source_pex_filepath, 0o775)
            else:
                deps_pex_filepaths.append(local_filepath)

        if not source_pex_filepath:
            raise ValueError("Invalid pex_tag has no source pex: %r" % pex_metadata.pex_tag)

        # we unpack each pex file into its own venv
        source_venv = self.venv_for(source_pex_filepath)
        deps_venvs = []
        for deps_filepath in deps_pex_filepaths:
            deps_venvs.append(self.venv_for(deps_filepath))

        entrypoint = str(source_venv.entrypoint)
        pythonpaths = ":".join(
            [str(source_venv.site_packages)]
            + [str(deps_venv.site_packages) for deps_venv in deps_venvs]
        )
        bin_paths = ":".join(
            [str(source_venv.bin)] + [str(deps_venv.bin) for deps_venv in deps_venvs]
        )
        working_dir = self.get_working_dir_for_pex(entrypoint)
        env = {
            "PATH": bin_paths + ":" + os.environ["PATH"],
            "PYTHONPATH": pythonpaths,
        }
        return PexExecutable(
            entrypoint,
            [source_pex_filepath] + deps_pex_filepaths,
            env,
            working_dir,
            venv_dirs=[str(source_venv.path)] + [str(deps_venv.path) for deps_venv in deps_venvs],
        )

    def venv_for(self, pex_filepath) -> PexVenv:
        _, pex_filename = os.path.split(pex_filepath)
        venv_dir = self.venv_dir_for(pex_filepath)
        if os.path.exists(venv_dir):
            logging.info("Reusing existing venv %r for %r", venv_dir, pex_filepath)
        else:
            self.install_venv(venv_dir, pex_filepath)
            if not os.path.exists(venv_dir):
                raise PexInstallationError("Could not install venv", pex_filepath)
        venv_path = Path(venv_dir).absolute()
        return PexVenv(
            path=venv_path,
            site_packages=self.get_site_packages_dir_for_venv(venv_path),
            bin=venv_path / "bin",
            entrypoint=venv_path / "pex",
            pex_filename=pex_filename,
        )

    def venv_dir_for(self, pex_filepath: str):
        # Use a short name for better stack traces
        short_hash = hashlib.shake_256(pex_filepath.encode("utf-8")).hexdigest(6)
        venv_root = os.getenv("VENVS_ROOT", "/venvs")
        return os.path.join(venv_root, short_hash)

    def install_venv(self, venv_dir: str, pex_filepath: str):
        # Unpacks the pex file into a venv at venv_dir
        try:
            subprocess.check_output(
                [
                    "pex-tools",
                    pex_filepath,
                    "venv",
                    # multiple packages sometimes provide the same file, eg dbt/__init__.py is in
                    # both dbt_core and dbt_duckdb
                    "--collisions-ok",
                    # since we combine multiple venvs, we need non hermetic scripts
                    "--non-hermetic-scripts",
                    venv_dir,
                ],
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            shutil.rmtree(venv_dir, ignore_errors=True)  # don't leave invalid dir behind
            raise PexInstallationError(
                f"Could not install venv. Pex output: {e.output}", pex_filepath
            ) from e
        logging.info(
            "Unpacked pex file %r into venv at %r",
            pex_filepath,
            venv_dir,
        )

    def get_site_packages_dir_for_venv(self, venv_path: Path) -> Path:
        python = venv_path / "bin" / "python3"
        proc = subprocess.run(
            [python, "-c", "import site; print(site.getsitepackages()[0])"],
            capture_output=True,
            check=False,
        )
        if not proc.returncode:
            return Path(proc.stdout.decode("utf-8").strip()).absolute()
        else:
            logging.error(
                "Cannot determine site-packages for venv at %r: %s\n%s",
                venv_path,
                proc.stdout.decode("utf-8"),
                proc.stderr.decode("utf-8"),
            )
            raise PexInstallationError("Cannot determine site-packages", venv_path, proc.stderr)

    def get_working_dir_for_pex(self, pex_path: str) -> Optional[str]:
        # A special 'working_directory' package may be included the source package.
        # If so we use site-packages/working_directory/root as the working dir.
        # This allows shipping arbitrary files to the server - also used for python_file support.
        if pex_path in self.working_dirs:
            return self.working_dirs[pex_path]
        try:
            working_dir_file = subprocess.check_output(
                [
                    pex_path,
                    "-c",
                    "import working_directory; print(working_directory.__file__);",
                ],
                encoding="utf-8",
            ).strip()
            if working_dir_file:
                package_dir, _ = working_dir_file.rsplit("/", 1)  # remove trailing __init__.py
                working_dir = os.path.join(package_dir, "root")
                self.working_dirs[pex_path] = working_dir
                return working_dir
            return None

        except subprocess.CalledProcessError:
            # working_directory package is optional, just log a message
            logging.info("Cannot import working_directory package - not setting current directory.")
            return None
        except OSError:
            # some issue with pex not being runnable, log an error but don't fail yet
            # might fail later if we try to run this again
            logging.exception(
                "Ignoring failure to run pex file to determine working_directory %r", pex_path
            )
            return None



import os
import subprocess
from typing import Optional

from dagster._serdes import deserialize_value
from dagster._utils.interrupts import capture_interrupts
from dagster_cloud_cli.core.workspace import PexMetadata
from typer import Option, Typer, echo

from ..registry import PexS3Registry
from ..server import run_multipex_server

app = Typer(hidden=True)


@app.command(short_help="Run multi-pex server to spin up subprocesses via PEX files")
def grpc(
    host: str = Option(default="localhost"),
    port: Optional[int] = Option(default=None),
    socket: Optional[str] = Option(default=None),
    max_workers: Optional[int] = Option(default=None),
    local_pex_files_dir: Optional[str] = Option(
        default="/tmp/pex-files", envvar="LOCAL_PEX_FILES_DIR"
    ),
    watchdog_run_interval: Optional[int] = Option(default=30, envvar="WATCHDOG_RUN_INTERVAL"),
):
    run_multipex_server(
        port=port,
        socket=socket,
        host=host,
        print_fn=echo,
        max_workers=max_workers,
        local_pex_files_dir=local_pex_files_dir,
        watchdog_run_interval=watchdog_run_interval,
    )


@app.command(short_help="Execute a run with a PEX file")
def execute_run(
    input_json: str,
    pex_metadata_json: str,
    local_pex_files_dir: Optional[str] = Option(
        default="/tmp/pex-files", envvar="LOCAL_PEX_FILES_DIR"
    ),
):
    with capture_interrupts():
        pex_metadata = deserialize_value(pex_metadata_json, PexMetadata)
        executable = PexS3Registry(local_pex_files_dir).get_pex_executable(pex_metadata)

        subprocess.run(
            [
                executable.source_path,
                "-m",
                "dagster",
                "api",
                "execute_run",
                input_json,
            ],
            check=True,
            env={**os.environ.copy(), **executable.environ},
            cwd=executable.working_directory,
        )



from dagster._core.storage.tags import HIDDEN_TAG_PREFIX, SYSTEM_TAG_PREFIX

CLOUD_SYSTEM_TAG_PREFIX = "dagster-cloud/"

PEX_METADATA_TAG = f"{HIDDEN_TAG_PREFIX}pex_metadata"
PEX_TAG_TAG = f"{SYSTEM_TAG_PREFIX}pex_tag"
